/*
 * Algebraic manipulator program.
 *
 * Mathomatic Copyright (c) 1996 George Gesslein II.
 *
 * Currently, this program only supports binary and unary operators.
 * Unary operators are implemented as a binary operation with a dummy operand.
 *
 * In the storage format, each level of parenthesis is indicated
 * by a level number (origin 1).  The deeper the level, the
 * higher the level number.
 *
 * The storage format for expressions is a fixed size array of elements
 * "token_type", which may be a CONSTANT, VARIABLE, or OPERATOR.
 * The array always alternates between operand (CONSTANT or VARIABLE)
 * and OPERATOR.  There is a separate integer for each array which
 * contains the current length of the expression stored in the array.
 * This length is always odd.
 *
 * Only one POWER operator is allowed per level in the storage format,
 * and no other operators may be on that level.  Same with the FACTORIAL
 * operator.
 *
 * Any number of TIMES and DIVIDE operators may be on the same
 * level, because they are simple multiplicative class operators.
 * The same for PLUS and MINUS.
 */

#include "am.h"
#include <signal.h>
#if	READLINE
#include <readline/readline.h>
#endif
#if	LINUX
#include <fenv.h>
#endif

#define	VERSION		"10.5"

void		fphandler();
void		inthandler();

int		n_equations;
int		cur_equation;

int		n_lhs[N_EQUATIONS];
int		n_rhs[N_EQUATIONS];

token_type	*lhs[N_EQUATIONS];
token_type	*rhs[N_EQUATIONS];

int		n_tlhs;
int		n_trhs;

token_type	tlhs[N_TOKENS];
token_type	trhs[N_TOKENS];

token_type	*scratch;

/* Set the following to "true" if you want upper and lower case variables. */
int		case_sensitive_flag = false;

/* Set the following to "true" to always display equations in fraction format. */
int		groupall = true;
int		screen_columns = DEFAULT_COLUMNS;
int		debug_level;
int		domain_check;
int		color_flag;	/* true for ANSI color mode */
int		html_flag;	/* true for HTML color mode */
int		partial_flag;	/* true for partial unfactoring */
int		symb_flag;	/* true for "simplify symbolic" */
int		high_prec;	/* true while saving equations */
int		input_column;
int		sign_flag;	/* true when all "sign" variables are to compare equal */
int		usage_flag;
char		*prog_name;
char		*tmp_file = "mathxxx.tmp";
double		epsilon;
char		var_str[80];	/* temp storage for variable names */
char		prompt_str[80];	/* temp storage for prompt strings */

sign_array_type	sign_array;

FILE		*gfp;

jmp_buf		jmp_save;

int		start;

int
main(argc, argv)
int	argc;
char	**argv;
{
	int		i;
	char		*cp;
	double		numerator, denominator;
	int		coption;

	prog_name = argv[0];
	init_gvars();
	gfp = stdout;
	setbuf(stdout, NULL);		/* make standard output unbuffered */
	epsilon = 0.00000000000005;	/* the correct value for doubles */
	start = 1;
	coption = false;
	if (argc > 1) {
		if (strcmp(argv[1], "-c") == 0) {
			start++;
			coption = true;
		} else if (strcmp(argv[1], "-h") == 0) {
			printf("<pre>\nHTML color mode enabled.\n");
			start++;
			html_flag = true;
		}
	}
	printf("Mathomatic Version %s\n", VERSION);
	print_copyright();
	init_mem();
	if (argc >= 1 && !html_flag) {
		i = strlen(argv[0]);
		for (; i > 0; i--) {
			if (argv[0][i] == '/' || argv[0][i] == '\\') {
				i = 0;
				break;
			}
			if (argv[0][i] == '.')
				break;
		}
		if (i <= 0) {
			i = strlen(argv[0]);
		}
		if (i && (argv[0][i-1] == 'C' || argv[0][i-1] == 'c')) {
			if (!coption) {
				printf("ANSI color mode enabled.  To disable, rename this program to \"am\",\n");
				printf("or use the \"-c\" option.\n");
			}
			color_flag = true;
		}
	}
	if (coption) {
		if (color_flag) {
			printf("ANSI color mode toggled off by the \"-c\" option.\n");
			color_flag = false;
		} else {
			printf("ANSI color mode enabled.\n");
			color_flag = true;
		}
	}
	printf("\n");
	default_color();
	if ((i = setjmp(jmp_save)) != 0) {
		clean_up();
		if (i != 3) {
			if (i == 13)
				printf("Operation abruptly aborted.\n");
			else
				printf("Operation aborted.\n");
		}
	} else {
#if	LINUX
		feenableexcept(FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW);
#endif
		signal(SIGFPE, fphandler);
		signal(SIGINT, inthandler);
		if (!f_to_fraction(0.5, &numerator, &denominator)
		    || !f_to_fraction(1.0/3.0, &numerator, &denominator)) {
			printf("Cannot convert floating point values to fractions.\n");
			printf("Your Floating Point Unit may be crippled.\n");
		}
#if	IBM_PC || UNIX
		for (i = start; i < argc; i++) {
			if (!read_in(argv[i])) {
				reedit(argv[i]);
				break;
			} else if (strcmp(argv[i], tmp_file) == 0) {
				unlink(tmp_file);
			}
		}
#else
		for (i = start; i < argc; i++) {
			if (!read_in(argv[i])) {
				break;
			}
		}
#endif
	}
	for (;;) {
		sprintf(prompt_str, "%d%s", cur_equation + 1, html_flag ? HTML_PROMPT : PROMPT);
		if ((cp = getstring((char *) &tlhs[0], N_TOKENS * sizeof(token_type))) == NULL)
			break;
		process(cp);
	}
	printf("\n");
	exit_program();
}

print_copyright()
{
	printf("Copyright (C) 1987-2003 George Gesslein II.\n");
}

/*
 * Allocate the equation spaces.
 */
init_mem()
{
	int	i;
	char	*junk;

	if ((junk = malloc(4096)) == NULL
	    || (scratch = (token_type *) malloc((N_TOKENS + 10) * sizeof(token_type))) == NULL) {
		printf("Not enough memory.\n");
		exit(1);
	}
	for (i = 0; i < N_EQUATIONS; i++) {
#if	!HALLOC
		lhs[i] = (token_type *) malloc(N_TOKENS * sizeof(token_type));
#else
		lhs[i] = (token_type *) halloc((long) N_TOKENS, sizeof(token_type));
#endif
		if (lhs[i] == NULL)
			break;
#if	!HALLOC
		rhs[i] = (token_type *) malloc(N_TOKENS * sizeof(token_type));
#else
		rhs[i] = (token_type *) halloc((long) N_TOKENS, sizeof(token_type));
#endif
		if (rhs[i] == NULL) {
#if	!HALLOC
			free((char *) lhs[i]);
#else
			hfree((char *) lhs[i]);
#endif
			break;
		}
	}
	free(junk);
	n_equations = i;
	clear("all");
	if (n_equations != N_EQUATIONS && n_equations < 5) {
		printf("\nDue to the small amount of available memory,\n");
		printf("you only have room for %d equations.\n", n_equations);
	} else {
		printf("%d equation spaces allocated (total size is %ld KBytes).\n",
		    n_equations, ((long) N_TOKENS * sizeof(token_type) * 2L * n_equations) / 1000L);
	}
}

#if	MSC
void
fphandler(sig, num)
int	sig, num;
{
	_fpreset();
	switch (num) {
	case FPE_INVALID:
		printf("Floating point value invalid!\n");
		break;
	case FPE_ZERODIVIDE:
		printf("Divide by zero error!\n");
		break;
	case FPE_OVERFLOW:
		printf("Floating point overflow!\n");
		break;
	case FPE_UNDERFLOW:
		printf("Floating point underflow!\n");
		break;
	case FPE_INEXACT:
		printf("Floating point result inexact!\n");
		break;
	default:
		printf("Unknown floating point error!  Code = 0x%x.\n", num);
		break;
	}
	longjmp(jmp_save, 2);
}
#else
void
fphandler(sig)
int	sig;
{
#if	LINUX
	printf("Floating point exception!\n");
	feenableexcept(FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW);
	signal(SIGFPE, fphandler);
#else
	printf("Floating point exception!\n");
	signal(SIGFPE, fphandler);
#endif
	longjmp(jmp_save, 2);
}
#endif

#if	MATHERR
int
matherr(x)
struct exception	*x;
{
	switch (x->type) {
	case DOMAIN:
		if (domain_check) {
			domain_check = false;
			return 1;
		}
		printf("Domain error in constant!\n");
		break;
	case SING:
		printf("Singularity error in constant!\n");
		break;
	case OVERFLOW:
		printf("Overflow error in constant!\n");
		break;
	case UNDERFLOW:
		printf("Underflow error in constant!\n");
		break;
	case TLOSS:
		printf("Total loss of significance in constant!\n");
		break;
	case PLOSS:
		printf("Partial loss of significance in constant!\n");
		break;
	default:
		printf("Unknown error in constant!\n");
		break;
	}
	longjmp(jmp_save, 2);
}

check_err()
{
}
#else
check_err()
{
	switch (errno) {
	case EDOM:
		errno = 0;
		if (domain_check) {
			domain_check = false;
		} else {
			printf("Domain error in constant!\n");
			longjmp(jmp_save, 2);
		}
		break;
	case ERANGE:
		errno = 0;
		printf("Overflow error in constant!\n");
		longjmp(jmp_save, 2);
		break;
	}
}
#endif

/*
 * Control-C handler.
 */
void
inthandler(sig, num)
int	sig, num;
{
	printf("\n");
	exit_program();
}

/*
 * Exit this program and return to the Operating System.
 */
exit_program()
{
	reset_attr();
	printf("Thank you for using Mathomatic!\n");
	printf("Please visit \"http://www.mathomatic.com\".\n");
	if (html_flag) {
		printf("</pre>\n");
	}
	exit(0);
}

/*
 * Common routine to output a prompt
 * and get a line of input from stdin.
 * All Mathomatic input comes from this routine,
 * except for file reading.
 *
 * Returns "string" if successful, NULL if error or end-of-file.
 */
char	*
getstring(string, n)
char	*string;	/* storage for input string */
int	n;		/* maximum size of "string" in bytes */
{
	char	*cp;

	input_column = strlen(prompt_str);
#if	READLINE
	cp = readline(prompt_str);
	if (cp == NULL)
		return NULL;
	strncpy(string, cp, n);
	string[n-1] = '\0';
	free(cp);
	if (string[0]) {
		add_history(string);
	}
	return string;
#else
	printf("%s", prompt_str);
	return(fgets(string, n, stdin));
#endif
}
